"use strict";
/*
  (c) 2012-2021 Noora Halme et al. (see AUTHORS)
  (c) 2022 Alasdair McLeay

  This code is licensed under the MIT license:
  http://www.opensource.org/licenses/mit-license.php

  Protracker module player class

  todo:
  - pattern looping is broken (see mod.black_queen)
  - properly test EEx delay pattern
*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.Protracker = void 0;
// constructor for protracker player object
var Protracker = /** @class */ (function () {
    function Protracker() {
        this.songlen = 0;
        this.patterntable = new Float32Array(0);
        this.channels = 0;
        this.sample = [];
        this.samples = 0;
        this.patterns = 0;
        this.pattern = [];
        this.looprow = 0;
        this.patternwait = 0;
        // used first in intialize
        this.tick = 0;
        this.position = 0;
        this.row = 0;
        this.offset = 0;
        this.flags = 0;
        this.speed = 0;
        this.bpm = 0;
        this.breakrow = 0;
        this.patternjump = 0;
        this.channel = [];
        // used first in parse
        this.chvu = new Float32Array(0);
        var t;
        this.clearsong();
        this.initialize();
        this.playing = false;
        this.paused = false;
        this.repeat = false;
        this.filter = false;
        this.mixval = 4.0;
        this.syncqueue = [];
        this.samplerate = 44100;
        // paula period values
        this.baseperiodtable = new Float32Array([
            856, 808, 762, 720, 678, 640, 604, 570, 538, 508, 480, 453, 428, 404, 381,
            360, 339, 320, 302, 285, 269, 254, 240, 226, 214, 202, 190, 180, 170, 160,
            151, 143, 135, 127, 120, 113,
        ]);
        // finetune multipliers
        this.finetunetable = new Float32Array(16);
        for (t = 0; t < 16; t++)
            this.finetunetable[t] = Math.pow(2, (t - 8) / 12 / 8);
        // calc tables for vibrato waveforms
        this.vibratotable = [];
        for (t = 0; t < 4; t++) {
            this.vibratotable[t] = new Float32Array(64);
            for (var i = 0; i < 64; i++) {
                switch (t) {
                    case 0:
                        this.vibratotable[t][i] = 127 * Math.sin(Math.PI * 2 * (i / 64));
                        break;
                    case 1:
                        this.vibratotable[t][i] = 127 - 4 * i;
                        break;
                    case 2:
                        this.vibratotable[t][i] = i < 32 ? 127 : -127;
                        break;
                    case 3:
                        this.vibratotable[t][i] = (1 - 2 * Math.random()) * 127;
                        break;
                }
            }
        }
    }
    // clear song data
    Protracker.prototype.clearsong = function () {
        this.title = "";
        this.signature = "";
        this.songlen = 1;
        this.repeatpos = 0;
        this.patterntable = new Float32Array(128);
        for (var i = 0; i < 128; i++)
            this.patterntable[i] = 0;
        this.channels = 4;
        this.sample = [];
        this.samples = 31;
        for (var i = 0; i < 31; i++) {
            this.sample[i] = {
                name: "",
                length: 0,
                finetune: 0,
                volume: 64,
                loopstart: 0,
                looplength: 0,
                data: new Float32Array(0),
            };
        }
        this.patterns = 0;
        this.pattern = [];
        this.note = [];
        this.pattern_unpack = [];
        this.looprow = 0;
        this.loopstart = 0;
        this.loopcount = 0;
        this.patterndelay = 0;
        this.patternwait = 0;
    };
    // initialize all player variables
    Protracker.prototype.initialize = function () {
        this.syncqueue = [];
        this.tick = 0;
        this.position = 0;
        this.row = 0;
        this.offset = 0;
        this.flags = 0;
        this.speed = 6;
        this.bpm = 125;
        this.breakrow = 0;
        this.patternjump = 0;
        this.patterndelay = 0;
        this.patternwait = 0;
        this.endofsong = false;
        this.channel = [];
        for (var i = 0; i < this.channels; i++) {
            this.channel[i] = {
                sample: 0,
                period: 214,
                voiceperiod: 214,
                note: 24,
                volume: 64,
                command: 0,
                data: 0,
                samplepos: 0,
                samplespeed: 0,
                flags: 0,
                noteon: 0,
                slidespeed: 0,
                slideto: 214,
                slidetospeed: 0,
                arpeggio: 0,
                semitone: 12,
                vibratospeed: 0,
                vibratodepth: 0,
                vibratopos: 0,
                vibratowave: 0,
            };
        }
    };
    // parse the module from local buffer
    Protracker.prototype.parse = function (buffer) {
        var i, j, c;
        for (i = 0; i < 4; i++)
            this.signature += String.fromCharCode(buffer[1080 + i]);
        switch (this.signature) {
            case "M.K.":
            case "M!K!":
            case "4CHN":
            case "FLT4":
                break;
            case "6CHN":
                this.channels = 6;
                break;
            case "8CHN":
            case "FLT8":
                this.channels = 8;
                break;
            case "28CH":
                this.channels = 28;
                break;
            default:
                return false;
        }
        this.chvu = new Float32Array();
        for (var i_1 = 0; i_1 < this.channels; i_1++)
            this.chvu[i_1] = 0.0;
        i = 0;
        while (buffer[i] && i < 20)
            this.title = this.title + String.fromCharCode(buffer[i++]);
        for (var i_2 = 0; i_2 < this.samples; i_2++) {
            var st = 20 + i_2 * 30;
            j = 0;
            while (buffer[st + j] && j < 22) {
                this.sample[i_2].name +=
                    buffer[st + j] > 0x1f && buffer[st + j] < 0x7f
                        ? String.fromCharCode(buffer[st + j])
                        : " ";
                j++;
            }
            this.sample[i_2].length = 2 * (buffer[st + 22] * 256 + buffer[st + 23]);
            this.sample[i_2].finetune = buffer[st + 24];
            if (this.sample[i_2].finetune > 7)
                this.sample[i_2].finetune = this.sample[i_2].finetune - 16;
            this.sample[i_2].volume = buffer[st + 25];
            this.sample[i_2].loopstart = 2 * (buffer[st + 26] * 256 + buffer[st + 27]);
            this.sample[i_2].looplength = 2 * (buffer[st + 28] * 256 + buffer[st + 29]);
            if (this.sample[i_2].looplength == 2)
                this.sample[i_2].looplength = 0;
            if (this.sample[i_2].loopstart > this.sample[i_2].length) {
                this.sample[i_2].loopstart = 0;
                this.sample[i_2].looplength = 0;
            }
        }
        this.songlen = buffer[950];
        if (buffer[951] != 127)
            this.repeatpos = buffer[951];
        for (var i_3 = 0; i_3 < 128; i_3++) {
            this.patterntable[i_3] = buffer[952 + i_3];
            if (this.patterntable[i_3] > this.patterns)
                this.patterns = this.patterntable[i_3];
        }
        this.patterns += 1;
        var patlen = 4 * 64 * this.channels;
        this.pattern = [];
        this.note = [];
        this.pattern_unpack = [];
        for (var i_4 = 0; i_4 < this.patterns; i_4++) {
            this.pattern[i_4] = new Uint8Array(patlen);
            this.note[i_4] = new Uint8Array(this.channels * 64);
            this.pattern_unpack[i_4] = new Uint8Array(this.channels * 64 * 5);
            for (j = 0; j < patlen; j++)
                this.pattern[i_4][j] = buffer[1084 + i_4 * patlen + j];
            for (j = 0; j < 64; j++)
                for (c = 0; c < this.channels; c++) {
                    this.note[i_4][j * this.channels + c] = 0;
                    var n = ((this.pattern[i_4][j * 4 * this.channels + c * 4] & 0x0f) << 8) |
                        this.pattern[i_4][j * 4 * this.channels + c * 4 + 1];
                    for (var np = 0; np < this.baseperiodtable.length; np++)
                        if (n == this.baseperiodtable[np])
                            this.note[i_4][j * this.channels + c] = np;
                }
            for (j = 0; j < 64; j++) {
                for (c = 0; c < this.channels; c++) {
                    var pp = j * 4 * this.channels + c * 4;
                    var ppu = j * 5 * this.channels + c * 5;
                    var n = ((this.pattern[i_4][pp] & 0x0f) << 8) | this.pattern[i_4][pp + 1];
                    if (n) {
                        n = this.note[i_4][j * this.channels + c];
                        n = n % 12 | ((Math.floor(n / 12) + 2) << 4);
                    }
                    this.pattern_unpack[i_4][ppu + 0] = n ? n : 255;
                    this.pattern_unpack[i_4][ppu + 1] =
                        (this.pattern[i_4][pp + 0] & 0xf0) | (this.pattern[i_4][pp + 2] >> 4);
                    this.pattern_unpack[i_4][ppu + 2] = 255;
                    this.pattern_unpack[i_4][ppu + 3] = this.pattern[i_4][pp + 2] & 0x0f;
                    this.pattern_unpack[i_4][ppu + 4] = this.pattern[i_4][pp + 3];
                }
            }
        }
        var sst = 1084 + this.patterns * patlen;
        for (var i_5 = 0; i_5 < this.samples; i_5++) {
            this.sample[i_5].data = new Float32Array(this.sample[i_5].length);
            for (j = 0; j < this.sample[i_5].length; j++) {
                var q = buffer[sst + j];
                if (q < 128) {
                    q = q / 128.0;
                }
                else {
                    q = (q - 128) / 128.0 - 1.0;
                }
                this.sample[i_5].data[j] = q;
            }
            sst += this.sample[i_5].length;
        }
        // look ahead at very first row to see if filter gets enabled
        this.filter = false;
        for (var ch = 0; ch < this.channels; ch++) {
            var p = this.patterntable[0];
            var pp = ch * 4;
            var cmd = this.pattern[p][pp + 2] & 0x0f, data = this.pattern[p][pp + 3];
            if (cmd == 0x0e && (data & 0xf0) == 0x00) {
                if (!(data & 0x01)) {
                    this.filter = true;
                }
                else {
                    this.filter = false;
                }
            }
        }
        // TODO: set lowpass cutoff
        // if (this.context) {
        //   if (this.filter) {
        //     this.lowpassNode.frequency.value = 3275;
        //   } else {
        //     this.lowpassNode.frequency.value = 28867;
        //   }
        // }
        this.chvu = new Float32Array(this.channels);
        for (var i_6 = 0; i_6 < this.channels; i_6++)
            this.chvu[i_6] = 0.0;
        return true;
    };
    Protracker.prototype.mix = function (bufs, buflen) {
        return mix(this, bufs, buflen);
    };
    Protracker.prototype.advance = function () {
        return advance(this);
    };
    Protracker.prototype.play = function () {
        this.playing = true;
        this.paused = false;
    };
    Protracker.prototype.pause = function () {
        this.playing = false;
        this.paused = true;
    };
    return Protracker;
}());
exports.Protracker = Protracker;
// advance player
function advance(mod) {
    var spd = (mod.samplerate * 60) / mod.bpm / 4 / 6;
    // advance player
    if (mod.offset > spd) {
        mod.tick++;
        mod.offset = 0;
        mod.flags |= 1;
    }
    if (mod.tick >= mod.speed) {
        if (mod.patterndelay) {
            // delay pattern
            if (mod.tick < (mod.patternwait + 1) * mod.speed) {
                mod.patternwait++;
            }
            else {
                mod.row++;
                mod.tick = 0;
                mod.flags |= 2;
                mod.patterndelay = 0;
            }
        }
        else {
            if (mod.flags & (16 + 32 + 64)) {
                if (mod.flags & 64) {
                    // loop pattern?
                    mod.row = mod.looprow;
                    mod.flags &= 0xa1;
                    mod.flags |= 2;
                }
                else {
                    if (mod.flags & 16) {
                        // pattern jump/break?
                        mod.position = mod.patternjump;
                        mod.row = mod.breakrow;
                        mod.flags &= 0xe1;
                        mod.flags |= 2;
                    }
                }
                mod.tick = 0;
            }
            else {
                mod.row++;
                mod.tick = 0;
                mod.flags |= 2;
            }
        }
    }
    if (mod.row >= 64) {
        mod.position++;
        mod.row = 0;
        mod.flags |= 4;
    }
    if (mod.position >= mod.songlen) {
        if (mod.repeat) {
            if (mod.breakrow || mod.patternjump) {
                mod.position = mod.patternjump;
                mod.row = mod.breakrow;
            }
            else {
                mod.position = 0;
            }
        }
        else {
            mod.endofsong = true;
            //mod.stop();
        }
        return;
    }
}
// mix an audio buffer with data
function mix(mod, bufs, buflen) {
    var f;
    var p, pp, n, nn;
    var outp = new Float32Array(2);
    for (var s = 0; s < buflen; s++) {
        outp[0] = 0.0;
        outp[1] = 0.0;
        if (!mod.paused && !mod.endofsong && mod.playing) {
            advance(mod);
            var och = 0;
            for (var ch = 0; ch < mod.channels; ch++) {
                // calculate playback position
                p = mod.patterntable[mod.position];
                pp = mod.row * 4 * mod.channels + ch * 4;
                if (mod.flags & 2) {
                    // new row
                    mod.channel[ch].command = mod.pattern[p][pp + 2] & 0x0f;
                    mod.channel[ch].data = mod.pattern[p][pp + 3];
                    if (!(mod.channel[ch].command == 0x0e &&
                        (mod.channel[ch].data & 0xf0) == 0xd0)) {
                        n = ((mod.pattern[p][pp] & 0x0f) << 8) | mod.pattern[p][pp + 1];
                        if (n) {
                            // noteon, except if command=3 (porta to note)
                            if (mod.channel[ch].command != 0x03 &&
                                mod.channel[ch].command != 0x05) {
                                mod.channel[ch].period = n;
                                mod.channel[ch].samplepos = 0;
                                if (mod.channel[ch].vibratowave > 3)
                                    mod.channel[ch].vibratopos = 0;
                                mod.channel[ch].flags |= 3; // recalc speed
                                mod.channel[ch].noteon = 1;
                            }
                            // in either case, set the slide to note target
                            mod.channel[ch].slideto = n;
                        }
                        nn =
                            (mod.pattern[p][pp + 0] & 0xf0) | (mod.pattern[p][pp + 2] >> 4);
                        if (nn) {
                            mod.channel[ch].sample = nn - 1;
                            mod.channel[ch].volume = mod.sample[nn - 1].volume;
                            if (!n && mod.channel[ch].samplepos > mod.sample[nn - 1].length)
                                mod.channel[ch].samplepos = 0;
                        }
                    }
                }
                mod.channel[ch].voiceperiod = mod.channel[ch].period;
                // kill empty samples
                if (!mod.sample[mod.channel[ch].sample].length)
                    mod.channel[ch].noteon = 0;
                // effects
                if (mod.flags & 1) {
                    if (!mod.tick) {
                        // process only on tick 0
                        effects_t0[mod.channel[ch].command](mod, ch);
                    }
                    else {
                        effects_t1[mod.channel[ch].command](mod, ch);
                    }
                }
                // recalc note number from period
                if (mod.channel[ch].flags & 2) {
                    for (var np = 0; np < mod.baseperiodtable.length; np++)
                        if (mod.baseperiodtable[np] >= mod.channel[ch].period)
                            mod.channel[ch].note = np;
                    mod.channel[ch].semitone = 7;
                    if (mod.channel[ch].period >= 120)
                        mod.channel[ch].semitone =
                            mod.baseperiodtable[mod.channel[ch].note] -
                                mod.baseperiodtable[mod.channel[ch].note + 1];
                }
                // recalc sample speed and apply finetune
                if ((mod.channel[ch].flags & 1 || mod.flags & 2) &&
                    mod.channel[ch].voiceperiod)
                    mod.channel[ch].samplespeed =
                        ((7093789.2 / (mod.channel[ch].voiceperiod * 2)) *
                            mod.finetunetable[mod.sample[mod.channel[ch].sample].finetune + 8]) /
                            mod.samplerate;
                // advance vibrato on each new tick
                if (mod.flags & 1) {
                    mod.channel[ch].vibratopos += mod.channel[ch].vibratospeed;
                    mod.channel[ch].vibratopos &= 0x3f;
                }
                // mix channel to output
                och = och ^ (ch & 1);
                f = 0.0;
                if (mod.channel[ch].noteon) {
                    if (mod.sample[mod.channel[ch].sample].length >
                        mod.channel[ch].samplepos)
                        f =
                            (mod.sample[mod.channel[ch].sample].data[Math.floor(mod.channel[ch].samplepos)] *
                                mod.channel[ch].volume) /
                                64.0;
                    outp[och] += f;
                    mod.channel[ch].samplepos += mod.channel[ch].samplespeed;
                }
                mod.chvu[ch] = Math.max(mod.chvu[ch], Math.abs(f));
                // loop or end samples
                if (mod.channel[ch].noteon) {
                    if (mod.sample[mod.channel[ch].sample].loopstart ||
                        mod.sample[mod.channel[ch].sample].looplength) {
                        if (mod.channel[ch].samplepos >=
                            mod.sample[mod.channel[ch].sample].loopstart +
                                mod.sample[mod.channel[ch].sample].looplength) {
                            mod.channel[ch].samplepos -=
                                mod.sample[mod.channel[ch].sample].looplength;
                        }
                    }
                    else {
                        if (mod.channel[ch].samplepos >=
                            mod.sample[mod.channel[ch].sample].length) {
                            mod.channel[ch].noteon = 0;
                        }
                    }
                }
                // clear channel flags
                mod.channel[ch].flags = 0;
            }
            mod.offset++;
            mod.flags &= 0x70;
        }
        // done - store to output buffer
        bufs[0][s] = outp[0];
        bufs[1][s] = outp[1];
    }
}
//
// tick 0 effect functions
//
var effect_t0_0 = function (mod, ch) {
    // 0 arpeggio
    mod.channel[ch].arpeggio = mod.channel[ch].data;
};
var effect_t0_1 = function (mod, ch) {
    // 1 slide up
    if (mod.channel[ch].data)
        mod.channel[ch].slidespeed = mod.channel[ch].data;
};
var effect_t0_2 = function (mod, ch) {
    // 2 slide down
    if (mod.channel[ch].data)
        mod.channel[ch].slidespeed = mod.channel[ch].data;
};
var effect_t0_3 = function (mod, ch) {
    // 3 slide to note
    if (mod.channel[ch].data)
        mod.channel[ch].slidetospeed = mod.channel[ch].data;
};
var effect_t0_4 = function (mod, ch) {
    // 4 vibrato
    if (mod.channel[ch].data & 0x0f && mod.channel[ch].data & 0xf0) {
        mod.channel[ch].vibratodepth = mod.channel[ch].data & 0x0f;
        mod.channel[ch].vibratospeed = (mod.channel[ch].data & 0xf0) >> 4;
    }
    effects_t1[4](mod, ch);
};
var effect_t0_5 = function (mod, ch) {
    // 5
};
var effect_t0_6 = function (mod, ch) {
    // 6
};
var effect_t0_7 = function (mod, ch) {
    // 7
};
var effect_t0_8 = function (mod, ch) {
    // 8 unused, used for syncing
    mod.syncqueue.unshift(mod.channel[ch].data & 0x0f);
};
var effect_t0_9 = function (mod, ch) {
    // 9 set sample offset
    mod.channel[ch].samplepos = mod.channel[ch].data * 256;
};
var effect_t0_a = function (mod, ch) {
    // a
};
var effect_t0_b = function (mod, ch) {
    // b pattern jump
    mod.breakrow = 0;
    mod.patternjump = mod.channel[ch].data;
    mod.flags |= 16;
};
var effect_t0_c = function (mod, ch) {
    // c set volume
    mod.channel[ch].volume = mod.channel[ch].data;
};
var effect_t0_d = function (mod, ch) {
    // d pattern break
    mod.breakrow =
        ((mod.channel[ch].data & 0xf0) >> 4) * 10 + (mod.channel[ch].data & 0x0f);
    if (!(mod.flags & 16))
        mod.patternjump = mod.position + 1;
    mod.flags |= 16;
};
var effect_t0_e = function (mod, ch) {
    // e
    var i = (mod.channel[ch].data & 0xf0) >> 4;
    effects_t0_e[i](mod, ch);
};
var effect_t0_f = function (mod, ch) {
    // f set speed
    if (mod.channel[ch].data > 32) {
        mod.bpm = mod.channel[ch].data;
    }
    else {
        if (mod.channel[ch].data)
            mod.speed = mod.channel[ch].data;
    }
};
//
// tick 0 effect e functions
//
var effect_t0_e0 = function (mod, ch) {
    // e0 filter on/off
    if (mod.channels > 4)
        return; // use only for 4ch amiga tunes
    if (mod.channel[ch].data & 0x01) {
        mod.filter = false;
    }
    else {
        mod.filter = true;
    }
};
var effect_t0_e1 = function (mod, ch) {
    // e1 fine slide up
    mod.channel[ch].period -= mod.channel[ch].data & 0x0f;
    if (mod.channel[ch].period < 113)
        mod.channel[ch].period = 113;
};
var effect_t0_e2 = function (mod, ch) {
    // e2 fine slide down
    mod.channel[ch].period += mod.channel[ch].data & 0x0f;
    if (mod.channel[ch].period > 856)
        mod.channel[ch].period = 856;
    mod.channel[ch].flags |= 1;
};
var effect_t0_e3 = function (mod, ch) {
    // e3 set glissando
};
var effect_t0_e4 = function (mod, ch) {
    // e4 set vibrato waveform
    mod.channel[ch].vibratowave = mod.channel[ch].data & 0x07;
};
var effect_t0_e5 = function (mod, ch) {
    // e5 set finetune
};
var effect_t0_e6 = function (mod, ch) {
    // e6 loop pattern
    if (mod.channel[ch].data & 0x0f) {
        if (mod.loopcount) {
            mod.loopcount--;
        }
        else {
            mod.loopcount = mod.channel[ch].data & 0x0f;
        }
        if (mod.loopcount)
            mod.flags |= 64;
    }
    else {
        mod.looprow = mod.row;
    }
};
var effect_t0_e7 = function (mod, ch) {
    // e7
};
var effect_t0_e8 = function (mod, ch) {
    // e8, use for syncing
    mod.syncqueue.unshift(mod.channel[ch].data & 0x0f);
};
var effect_t0_e9 = function (mod, ch) {
    // e9
};
var effect_t0_ea = function (mod, ch) {
    // ea fine volslide up
    mod.channel[ch].volume += mod.channel[ch].data & 0x0f;
    if (mod.channel[ch].volume > 64)
        mod.channel[ch].volume = 64;
};
var effect_t0_eb = function (mod, ch) {
    // eb fine volslide down
    mod.channel[ch].volume -= mod.channel[ch].data & 0x0f;
    if (mod.channel[ch].volume < 0)
        mod.channel[ch].volume = 0;
};
var effect_t0_ec = function (mod, ch) {
    // ec
};
var effect_t0_ed = function (mod, ch) {
    // ed delay sample
    if (mod.tick == (mod.channel[ch].data & 0x0f)) {
        // start note
        var p = mod.patterntable[mod.position];
        var pp = mod.row * 4 * mod.channels + ch * 4;
        var n = ((mod.pattern[p][pp] & 0x0f) << 8) | mod.pattern[p][pp + 1];
        if (n) {
            mod.channel[ch].period = n;
            mod.channel[ch].voiceperiod = mod.channel[ch].period;
            mod.channel[ch].samplepos = 0;
            if (mod.channel[ch].vibratowave > 3)
                mod.channel[ch].vibratopos = 0;
            mod.channel[ch].flags |= 3; // recalc speed
            mod.channel[ch].noteon = 1;
        }
        n = (mod.pattern[p][pp + 0] & 0xf0) | (mod.pattern[p][pp + 2] >> 4);
        if (n) {
            mod.channel[ch].sample = n - 1;
            mod.channel[ch].volume = mod.sample[n - 1].volume;
        }
    }
};
var effect_t0_ee = function (mod, ch) {
    // ee delay pattern
    mod.patterndelay = mod.channel[ch].data & 0x0f;
    mod.patternwait = 0;
};
var effect_t0_ef = function (mod, ch) {
    // ef
};
//
// tick 1+ effect functions
//
var effect_t1_0 = function (mod, ch) {
    // 0 arpeggio
    if (mod.channel[ch].data) {
        var apn = mod.channel[ch].note;
        if (mod.tick % 3 == 1)
            apn += mod.channel[ch].arpeggio >> 4;
        if (mod.tick % 3 == 2)
            apn += mod.channel[ch].arpeggio & 0x0f;
        if (apn >= 0 && apn <= mod.baseperiodtable.length)
            mod.channel[ch].voiceperiod = mod.baseperiodtable[apn];
        mod.channel[ch].flags |= 1;
    }
};
var effect_t1_1 = function (mod, ch) {
    // 1 slide up
    mod.channel[ch].period -= mod.channel[ch].slidespeed;
    if (mod.channel[ch].period < 113)
        mod.channel[ch].period = 113;
    mod.channel[ch].flags |= 3; // recalc speed
};
var effect_t1_2 = function (mod, ch) {
    // 2 slide down
    mod.channel[ch].period += mod.channel[ch].slidespeed;
    if (mod.channel[ch].period > 856)
        mod.channel[ch].period = 856;
    mod.channel[ch].flags |= 3; // recalc speed
};
var effect_t1_3 = function (mod, ch) {
    // 3 slide to note
    if (mod.channel[ch].period < mod.channel[ch].slideto) {
        mod.channel[ch].period += mod.channel[ch].slidetospeed;
        if (mod.channel[ch].period > mod.channel[ch].slideto)
            mod.channel[ch].period = mod.channel[ch].slideto;
    }
    if (mod.channel[ch].period > mod.channel[ch].slideto) {
        mod.channel[ch].period -= mod.channel[ch].slidetospeed;
        if (mod.channel[ch].period < mod.channel[ch].slideto)
            mod.channel[ch].period = mod.channel[ch].slideto;
    }
    mod.channel[ch].flags |= 3; // recalc speed
};
var effect_t1_4 = function (mod, ch) {
    // 4 vibrato
    var waveform = mod.vibratotable[mod.channel[ch].vibratowave & 3][mod.channel[ch].vibratopos] / 63.0; //127.0;
    // two different implementations for vibrato
    //  var a=(mod.channel[ch].vibratodepth/32)*mod.channel[ch].semitone*waveform; // non-linear vibrato +/- semitone
    var a = mod.channel[ch].vibratodepth * waveform; // linear vibrato, depth has more effect high notes
    mod.channel[ch].voiceperiod += a;
    mod.channel[ch].flags |= 1;
};
var effect_t1_5 = function (mod, ch) {
    // 5 volslide + slide to note
    effect_t1_3(mod, ch); // slide to note
    effect_t1_a(mod, ch); // volslide
};
var effect_t1_6 = function (mod, ch) {
    // 6 volslide + vibrato
    effect_t1_4(mod, ch); // vibrato
    effect_t1_a(mod, ch); // volslide
};
var effect_t1_7 = function (mod, ch) {
    // 7
};
var effect_t1_8 = function (mod, ch) {
    // 8 unused
};
var effect_t1_9 = function (mod, ch) {
    // 9 set sample offset
};
var effect_t1_a = function (mod, ch) {
    // a volume slide
    if (!(mod.channel[ch].data & 0x0f)) {
        // y is zero, slide up
        mod.channel[ch].volume += mod.channel[ch].data >> 4;
        if (mod.channel[ch].volume > 64)
            mod.channel[ch].volume = 64;
    }
    if (!(mod.channel[ch].data & 0xf0)) {
        // x is zero, slide down
        mod.channel[ch].volume -= mod.channel[ch].data & 0x0f;
        if (mod.channel[ch].volume < 0)
            mod.channel[ch].volume = 0;
    }
};
var effect_t1_b = function (mod, ch) {
    // b pattern jump
};
var effect_t1_c = function (mod, ch) {
    // c set volume
};
var effect_t1_d = function (mod, ch) {
    // d pattern break
};
var effect_t1_e = function (mod, ch) {
    // e
    var i = (mod.channel[ch].data & 0xf0) >> 4;
    effects_t1_e[i](mod, ch);
};
var effect_t1_f = function (mod, ch) {
    // f
};
//
// tick 1+ effect e functions
//
var effect_t1_e0 = function (mod, ch) {
    // e0
};
var effect_t1_e1 = function (mod, ch) {
    // e1
};
var effect_t1_e2 = function (mod, ch) {
    // e2
};
var effect_t1_e3 = function (mod, ch) {
    // e3
};
var effect_t1_e4 = function (mod, ch) {
    // e4
};
var effect_t1_e5 = function (mod, ch) {
    // e5
};
var effect_t1_e6 = function (mod, ch) {
    // e6
};
var effect_t1_e7 = function (mod, ch) {
    // e7
};
var effect_t1_e8 = function (mod, ch) {
    // e8
};
var effect_t1_e9 = function (mod, ch) {
    // e9 retrig sample
    if (mod.tick % (mod.channel[ch].data & 0x0f) == 0)
        mod.channel[ch].samplepos = 0;
};
var effect_t1_ea = function (mod, ch) {
    // ea
};
var effect_t1_eb = function (mod, ch) {
    // eb
};
var effect_t1_ec = function (mod, ch) {
    // ec cut sample
    if (mod.tick == (mod.channel[ch].data & 0x0f))
        mod.channel[ch].volume = 0;
};
var effect_t1_ed = function (mod, ch) {
    // ed delay sample
    effect_t0_ed(mod, ch);
};
var effect_t1_ee = function (mod, ch) {
    // ee
};
var effect_t1_ef = function (mod, ch) {
    // ef
};
// effect jumptables
var effects_t0 = [
    effect_t0_0,
    effect_t0_1,
    effect_t0_2,
    effect_t0_3,
    effect_t0_4,
    effect_t0_5,
    effect_t0_6,
    effect_t0_7,
    effect_t0_8,
    effect_t0_9,
    effect_t0_a,
    effect_t0_b,
    effect_t0_c,
    effect_t0_d,
    effect_t0_e,
    effect_t0_f,
];
var effects_t0_e = [
    effect_t0_e0,
    effect_t0_e1,
    effect_t0_e2,
    effect_t0_e3,
    effect_t0_e4,
    effect_t0_e5,
    effect_t0_e6,
    effect_t0_e7,
    effect_t0_e8,
    effect_t0_e9,
    effect_t0_ea,
    effect_t0_eb,
    effect_t0_ec,
    effect_t0_ed,
    effect_t0_ee,
    effect_t0_ef,
];
var effects_t1 = [
    effect_t1_0,
    effect_t1_1,
    effect_t1_2,
    effect_t1_3,
    effect_t1_4,
    effect_t1_5,
    effect_t1_6,
    effect_t1_7,
    effect_t1_8,
    effect_t1_9,
    effect_t1_a,
    effect_t1_b,
    effect_t1_c,
    effect_t1_d,
    effect_t1_e,
    effect_t1_f,
];
var effects_t1_e = [
    effect_t1_e0,
    effect_t1_e1,
    effect_t1_e2,
    effect_t1_e3,
    effect_t1_e4,
    effect_t1_e5,
    effect_t1_e6,
    effect_t1_e7,
    effect_t1_e8,
    effect_t1_e9,
    effect_t1_ea,
    effect_t1_eb,
    effect_t1_ec,
    effect_t1_ed,
    effect_t1_ee,
    effect_t1_ef,
];
