"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.loadUrl = void 0;
var Protracker_1 = require("./Protracker");
var supportedformats = ["mod", "s3m", "xm"];
function loadUrl(url, options) {
    var _a, _b, _c;
    var ext = (_a = url.split(".").pop()) === null || _a === void 0 ? void 0 : _a.toLowerCase().trim();
    if (!ext || supportedformats.indexOf(ext) == -1) {
        // unknown extension, maybe amiga-style prefix?
        ext = (_c = (_b = url.split("/").pop()) === null || _b === void 0 ? void 0 : _b.split(".").shift()) === null || _c === void 0 ? void 0 : _c.toLowerCase().trim();
        if (!ext || supportedformats.indexOf(ext) == -1) {
            // ok, give up
            throw new Error("Unsupported file extension ".concat(ext));
        }
    }
    var player;
    switch (ext) {
        case "mod":
            player = new Protracker_1.Protracker();
            break;
        // case "s3m":
        //   player = new Screamtracker();
        //   break;
        // case "xm":
        //   player = new Fasttracker();
        //   break;
        default:
            throw new Error("Unsupported file extension ".concat(ext));
    }
    createContext(player, {
        context: options === null || options === void 0 ? void 0 : options.context,
        destination: options === null || options === void 0 ? void 0 : options.destination,
    });
    var request = new XMLHttpRequest();
    request.open("GET", url, true);
    request.responseType = "arraybuffer";
    request.onload = function () {
        var buffer = new Uint8Array(request.response);
        if (player.parse(buffer)) {
            player.initialize();
            player.flags = 1 + 2;
            if (options === null || options === void 0 ? void 0 : options.repeat) {
                player.repeat = true;
            }
            if (options === null || options === void 0 ? void 0 : options.autoplay) {
                player.play();
            }
        }
        else {
            throw new Error("Unable to parse buffer");
        }
    };
    request.send();
    return player;
}
exports.loadUrl = loadUrl;
function createContext(player, options) {
    var context = (options === null || options === void 0 ? void 0 : options.context) || new AudioContext();
    var samplerate = context.sampleRate;
    var bufferlen = samplerate > 44100 ? 4096 : 2048;
    var filterNode = context.createBiquadFilter();
    filterNode.frequency.value = 22050;
    // "LED filter" at 3275kHz - off by default
    // TODO:
    // const lowpassNode = context.createBiquadFilter();
    // player.setfilter(filter);
    // mixer
    var mixerNode = context.createScriptProcessor(bufferlen, 1, 2);
    mixerNode.onaudioprocess = function (event) {
        var bufs = [
            event.outputBuffer.getChannelData(0),
            event.outputBuffer.getChannelData(1),
        ];
        var buflen = event.outputBuffer.length;
        player.mix(bufs, buflen);
    };
    mixerNode.connect(filterNode);
    filterNode.connect((options === null || options === void 0 ? void 0 : options.destination) || context.destination);
    // lowpassNode.connect(context.destination);
    return context;
}
